/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TOPOLOGICAL_MAP_EXCEPTION_H
#define TOPOLOGICAL_MAP_EXCEPTION_H

#include <boost/format.hpp>
#include <stdexcept>
#include <topological_map/grid_graph.h>
#include <topological_map/roadmap.h>
#include <topological_map/point_ops.h>
#include <costmap_2d/costmap_2d.h>


namespace topological_map
{

using boost::format;
using std::string;

/// \brief A base class for all topological_map exceptions_ 
class TopologicalMapException: public std::runtime_error
{ 
public:
  TopologicalMapException(const format& error_string) : std::runtime_error(error_string.str()) {};
  TopologicalMapException(const char* str) : std::runtime_error(str) {};
};

/// \brief Exception denoting unknown grid cell
struct UnknownGridCellException: public TopologicalMapException
{
  UnknownGridCellException(const Cell2D& p) : TopologicalMapException(format("Unknown grid cell %1%") % p), cell(p) {}
  Cell2D cell;
};

/// \brief Exception denoting unknown 2d point
class UnknownPointException: public TopologicalMapException
{
public:
  UnknownPointException (const double x, const double y) : TopologicalMapException(format("Illegal 2d point (%1%, %2%)") % x % y) {}
  UnknownPointException (const double x, const double y, const double xmax, const double ymax) : 
    TopologicalMapException(format("Illegal 2d point (%1%, %2%) (should be nonnegative and < (%3%, %4%))") % x % y % xmax % ymax) {}
};

/// \brief Exception denoting point not on costmap
class UnknownCostmapPointException: public TopologicalMapException
{
public:
  UnknownCostmapPointException (const costmap_2d::Costmap2D& cmap, const geometry_msgs::Point& p):
    TopologicalMapException(format("Point %1%, %2%, %3% not on costmap") % p.x % p.y % p.z) {}
};

/// \brief Exception for unknown node id
struct UnknownNodeIdException: public TopologicalMapException
{
  UnknownNodeIdException (NodeId id) : TopologicalMapException(format("Unknown node id %1%") % id), id(id) {}
  const NodeId id;
};


/// \brief Exception for not being able to open top map file
class FileOpenException: public TopologicalMapException
{
public: FileOpenException (const string& filename) : TopologicalMapException(format("Unable to open file %1%") % filename) {}
};

/// \brief Exception for not being able to parse top map file
class FileFormatException: public TopologicalMapException
{
public: 
  FileFormatException (const string& filename): TopologicalMapException(format("File %1% was not correctly formatted") % filename) {}
  
};

/// \brief Exception for not finding a path in the roadmap
class NoPathFoundException: public TopologicalMapException
{
public: NoPathFoundException (NodeId i, NodeId j) : TopologicalMapException(format("No path found in roadmap between nodes %1% and %2%") % i % j) {}
};

/// \brief Exception thrown when calling goalDistance without setting goal
class GoalNotSetException: public TopologicalMapException
{
public: GoalNotSetException() : TopologicalMapException(format("Called goalDistance without setting goal first")) {}
};

struct XmlException: public TopologicalMapException
{
  XmlException(const string& filename, const char* what) : TopologicalMapException(format("Error reading xml file %1% : %2%") % filename % what) {}
  XmlException(const string& filename, const format& what) : TopologicalMapException(format("Error reading xml file %1% : %2%") % filename % what) {}
};

struct XmlRootException: public XmlException
{
  XmlRootException(const string& filename, const string& name, const string& expected) : XmlException(filename, format("Root name was %1% instead of %2%") % name % expected) {}
};

struct GridFileTypeException: public TopologicalMapException
{
  GridFileTypeException(const string& type) : TopologicalMapException(format("Type of file was %1% instead of P5") % type) {}
};

struct ObservationOutOfSequenceException: public TopologicalMapException
{
  ObservationOutOfSequenceException (const ros::Time& stamp1, const ros::Time& stamp2) : 
    TopologicalMapException (format ("Observations at %1% and %2% are out of sequence") % stamp1 % stamp2) {}
};

struct DuplicateNodeIdException: public TopologicalMapException
{
  DuplicateNodeIdException (const NodeId id) :
    TopologicalMapException (format ("Node id %1% already exists in roadmap") % id) {}
};

struct DuplicateEdgeException: public TopologicalMapException
{
  DuplicateEdgeException (const NodeId i, const NodeId j) :
    TopologicalMapException (format ("An edge between %1% and %2% already exists") % i % j) {}
};

struct SelfLoopException: public TopologicalMapException
{
  SelfLoopException (const NodeId i) :
    TopologicalMapException (format ("Tried to add a self-loop edge at node %1%") % i) {}
};

struct NonexistentEdgeException: public TopologicalMapException
{
  NonexistentEdgeException (const NodeId i, const NodeId j) :
    TopologicalMapException (format ("No edge in graph between nodes %1% and %2%") % i % j) {}
};

} // namespace topological_map


#endif
