/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file 
 * 
 * Incrementally computes a roadmap in response to a possibly changing static map
 */

#ifndef TOPOLOGICAL_MAP_INCREMENTAL_ROADMAP_H
#define TOPOLOGICAL_MAP_INCREMENTAL_ROADMAP_H

#include <topological_map/roadmap.h>
#include <topological_map/roadmap_msg.h>
#include <topological_map/grid_graph.h>
#include <costmap_2d/costmap_2d.h>

namespace topological_map
{

using std::set;

/// Governs how far from an initial candidate waypoint we search for a better one.  Must be between 0 and 1.
const double WAYPOINT_SEARCH_MULTIPLIER = .5;

/// Multiplied by edge_distance_threshold to determine max allowed edge length
const double EDGE_DISTANCE_MULTIPLIER = 3.0;

/// Represents state of incremental roadmap server
/// Does not have any internal logic pertaining to threads or locking.  Instead, when using 
/// an object of this class, const methods must be called holding (at least) a shared lock, 
/// and non-const methods must be called holding an exclusive lock.
class IncrementalRoadmap
{
public:

/************************************************************
 * Constructor
 ************************************************************/

  /// \brief Constructor 
  /// \param cmap Initial costmap
  /// \param edge_distance_threshold Map will strive to make every point within this distance of a waypoint
  IncrementalRoadmap (const costmap_2d::Costmap2D& cmap, double edge_distance_threshold);


/************************************************************
 * Const methods
 ************************************************************/

  /// Get snapshot of current roadmap.  Note that unless you copy the return value,
  /// you need to hold a shared lock while looking at it.
  const Roadmap& roadmapSnapshot () const;

  /// Find the cell near \a c that is furthest from obstacles
  /// "near" means within shortest-path radius edge_distance_threshold_*WAYPOINT_SEARCH_MULTIPLIER
  Cell2D bestNearbyCell (const Cell2D& c) const;

  /// Return whether a given cell is covered
  bool isCovered (const Cell2D& c) const;

  /// Return a number that expresses how good it is to prune a given edge, i.e., how much
  /// the shortest path distance goes up if the edge is removed 
  /// (if it becomes infinite, return 10*the edge cost)
  ///
  /// param roadmap_copy must have been obtained via a call to roadmapSnapshot(), and no 
  /// non-const operations must have happened to it or this object since then.
  double edgePruneCost (NodeId id, NodeId id2, Roadmap* roadmap_copy) const;

  /// Returns true iff \a cell is on the map
  bool knownFree (const Cell2D& cell) const;



/************************************************************
 * non-const methods
 ************************************************************/

  /// Update the metric map.  
  /// \throws TopologicalMapException if the new map doesn't have the right dimensions
  void updateMetricMap (const costmap_2d::Costmap2D& cmap);

  /// \brief Add a node to the roadmap and connect it to nearby nodes
  /// \return Id of new node
  NodeId addNode (const Cell2D& c);

  /// Remove an edge from the roadmap
  void removeEdge (NodeId n1, NodeId n2);

  // Set the roadmap
  void setRoadmap (const Roadmap& roadmap);
  
/************************************************************
 * Field accessors
 ************************************************************/

  /// get the x size of the metric map
  unsigned getXSize() const;

  /// get the y size of the metric map
  unsigned getYSize() const;

  /// get the resolution of the metric map
  double getResolution () const;

private:

  /************************************************************
   * const
   ************************************************************/

  // Compute and return distances from each cell to nearest waypoint
  multi_array<unsigned, 2> computeWaypointDistances () const;

  // Get set of node cells
  set<Cell2D> nodeCells () const;

  // Repropagate waypoint distances from a set of nodes
  void repropagateWaypointDistances(multi_array<unsigned, 2>* distances, const set<Cell2D>& waypoints) const;

  // Get nearby cells to a given cell (and distances to them)
  map<Cell2D, unsigned> cellsWithinRadius (const Cell2D& cell, const unsigned r) const;


/************************************************************ 
 * non-const 
 ************************************************************/

  /// Connect node \a id to nearby nodes
  void connectToNearbyNodes (NodeId id);

  /// Add an edge in the roadmap
  void addEdge (NodeId n1, NodeId n2, double length);

 

/************************************************************
 * Parameters
 ************************************************************/
  
  // Map parameters
  const vector<unsigned> dims_;
  const double resolution_;

  // Every point should be within this distance of a waypoint
  const double edge_distance_threshold_;
  
  // Grid equivalent of edge distance threshold
  const unsigned max_waypoint_dist_;


/************************************************************
 * State
 * The field descriptions below are guaranteed to hold except 
 * during non-const operations
 ************************************************************/

  // Current roadmap
  Roadmap roadmap_;

  // Last received metric map
  costmap_2d::Costmap2D metric_map_;

  // Grid graph encoding connectivity of metric_map_
  GridGraph grid_graph_;
  
  // Manhattan distances from each cell to nearest obstacle in metric_map_
  multi_array<unsigned, 2> obstacle_distances_;

  // Shortest path distances in grid_graph_ from each cell to a waypoint
  multi_array<unsigned, 2> waypoint_distances_;

};

} // topological_map

#endif // include guard
