/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file 
 * 
 * Data structure for a roadmap over points
 *
 * \author Bhaskara Marthi
 */


#ifndef TOPOLOGICAL_MAP_ROADMAP_H
#define TOPOLOGICAL_MAP_ROADMAP_H

#include <topological_map/region.h>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/optional.hpp>
#include <map>
#include <vector>
#include <set>
#include <iostream>
#include <geometry_msgs/Point.h>

namespace topological_map {

using boost::listS;
using boost::undirectedS;
using boost::adjacency_list;
using boost::graph_traits;
using std::map;
using std::vector;
using std::set;
using std::pair;
using std::istream;
using std::ostream;

typedef unsigned NodeId;


/************************************************************
 * Internally used definitions for the graph
 ************************************************************/


struct NodeInfo
{
  NodeInfo(NodeId id, const geometry_msgs::Point& point) : id(id), point(point) {}
  NodeId id; // Stable id 
  unsigned index; // Not stable; used internally by dijkstra search
  geometry_msgs::Point point;
};

struct EdgeInfo
{
  double cost;
};

typedef adjacency_list<listS, listS, undirectedS, NodeInfo, EdgeInfo> RoadmapImpl;
typedef graph_traits<RoadmapImpl>::vertex_descriptor RoadmapVertex;
typedef graph_traits<RoadmapImpl>::edge_descriptor RoadmapEdge;
typedef graph_traits<RoadmapImpl>::adjacency_iterator RoadmapAdjacencyIterator;
typedef graph_traits<RoadmapImpl>::vertex_iterator VertexIterator;
typedef graph_traits<RoadmapImpl>::edge_iterator RoadmapEdgeIterator;
typedef map<NodeId, RoadmapVertex> NodeIdVertexMap;
typedef pair<NodeId, double> NodeCost;


/// \brief Represents a topological roadmap over points
/// 
/// Provides the standard graph operations and shortest path computations.  Nodes have
/// ids, of type NodeId (which is an unsigned integer type) and edges are referred to
/// by their adjacent nodes.  Id's are stable in that a given id will never refer
/// to two different points (it may stop referring to a node if that node is removed
/// from the graph).
class Roadmap
{
public:

  /// \brief Construct empty roadmap
  Roadmap() : next_id_(1) {}

  /// \brief Copy constructor
  Roadmap (const Roadmap&);

  /// \brief Assignment operator
  Roadmap& operator=(const Roadmap& r);

  /// \brief Add a node without specifying the id
  /// \retval the new node's id
  NodeId addNode (const geometry_msgs::Point& p);

  /// \brief Add a node at point \a p with id \a id
  /// \throws DuplicateNodeIdException if there's already a node with this id
  void addNode (const geometry_msgs::Point& p, NodeId id);

  /// \brief Add an edge
  /// \post There exists a single edge between nodes \a i and \a j with cost \a cost
  /// \throws DuplicateEdgeException if there's already an edge between these nodes
  /// \throws SelfLoopException if i == j
  void addEdge (NodeId i, NodeId j, double cost);

  /// \brief Get the cost of an edge
  /// \retval Cost of the edge between nodes \a i and \a j
  /// \throws NonexistentEdgeException if there's no edge
  double getCost (const NodeId i, const NodeId j) const;

  /// \brief Remove a node from the graph
  /// \post Node \a i and all adjacent edges have been removed from the graph
  /// \throws UnknownNodeIdException
  void removeNode (NodeId i);

  /// \brief Remove edge from roadmap
  /// \post The edge between \a i and \a j has been removed from the graph
  /// \throws UnknownNodeIdException
  /// \throws NonexistentEdgeException
  void removeEdge (const NodeId i, const NodeId j);

  /// \brief Return the point associated to this node
  geometry_msgs::Point nodePoint (const NodeId id) const;
  
  /// \brief Shortest path cost in graph between two nodes
  /// \retval If a path exists, the cost of the minimal path; otherwise it's unset
  boost::optional<double> costBetween (NodeId i, NodeId j);

  /// \brief Shortest path between pair of nodes
  /// \retval The shortest path, if it exists, or unset otherwise
  boost::optional<vector<NodeId> > shortestPath (NodeId i, NodeId j);

  /// \brief For each adjacent node k to i, return cost of path from i to j starting by going to k
  /// \retval Vector of pairs, each consisting of a node id k and a double cost
  vector<NodeCost> nodeCosts (const NodeId i, const NodeId j);

  /// \brief Get the set of node ids in graph
  set<NodeId> allNodes () const;

  /// \brief Get the set of neighbor node ids of a node
  set<NodeId> neighbors (NodeId id) const;

  /// \brief Get nodes within a given radius (in Euclidean distance) of point
  set<NodeId> nodesWithinRadius (geometry_msgs::Point p, double radius) const;

private:

  RoadmapEdge ensureEdge(RoadmapVertex v, RoadmapVertex w);
  RoadmapVertex idVertex(NodeId i) const;
  bool idExists(NodeId i) const;
  void resetIndices();

  NodeId next_id_;
  NodeIdVertexMap id_vertex_map_;
  RoadmapImpl graph_;

};


/// \brief RAII object for a single temporary roadmap node.
/// Node and connected edges will be removed when this object is destroyed.
class TemporaryRoadmapNode
{
public:
  TemporaryRoadmapNode (Roadmap* roadmap, const geometry_msgs::Point& p);
  ~TemporaryRoadmapNode ();
  NodeId getId() const;

private:
  Roadmap* roadmap_;
  const NodeId id_;
};



} // namespace

#endif



