/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

#include <topological_map/roadmap_msg.h>
#include <topological_map/GetTopologicalPlan.h>
#include <ros/ros.h>
#include <boost/optional.hpp>
#include <boost/foreach.hpp>
#include <algorithm>
#include <set>

namespace topological_map {

using std::swap;
using std::set;
using std::vector;

typedef boost::optional<vector<NodeId> > PlanningResult;
typedef std::pair<NodeId, NodeId> NodePair;

class DeterministicTopoPlanner
{
public:
  DeterministicTopoPlanner ();

private:

  /****************************************
   * Callbacks
   ****************************************/
  
  bool makePlan (GetTopologicalPlan::Request& req, GetTopologicalPlan::Response& resp);

  /****************************************
   * Ops
   ****************************************/
  
  bool isBlocked (NodeId n1, NodeId n2) const;
  void observeBlocked (NodeId n1, NodeId n2);


  /****************************************
   * Params
   ****************************************/



  /****************************************
   * State
   ****************************************/
  
  set<NodePair> blocked_edges_;

  /****************************************
   * Associated objects
   ****************************************/

  ros::NodeHandle nh_;
  ros::ServiceServer topo_plan_server_;

};


/************************************************************
 * Initialization
 ************************************************************/

DeterministicTopoPlanner::DeterministicTopoPlanner () :
  topo_plan_server_(nh_.advertiseService("get_topological_plan", 
                                         &DeterministicTopoPlanner::makePlan, this))
{}



/************************************************************
 * Blocked edges
 ************************************************************/

bool DeterministicTopoPlanner::isBlocked (NodeId n1, NodeId n2) const
{
  if (n1>n2) 
    swap(n1, n2);
  
  NodePair e(n1, n2);
  return (blocked_edges_.find(e)!=blocked_edges_.end());
}

void DeterministicTopoPlanner::observeBlocked (NodeId n1, NodeId n2) 
{
  if (n1>n2)
    swap(n1, n2);
  NodePair e(n1, n2);
  ROS_DEBUG_STREAM_NAMED ("det_topo_planner", "Observing " << e.first << "--" << e.second << " is blocked");
  blocked_edges_.insert(e);
}


/************************************************************
 * Planner
 ************************************************************/

bool DeterministicTopoPlanner::makePlan (GetTopologicalPlan::Request& req, GetTopologicalPlan::Response& resp)
{
  Roadmap roadmap = message::roadmapFromRos(req.graph);
  ROS_DEBUG_STREAM_NAMED ("det_topo_planner", "Received planning request from "
                          << req.pos << " to " << req.goal << " in graph with "
                          << roadmap.allNodes().size() << " nodes");
  // Add new blocked edges
  BOOST_FOREACH (Edge e, req.blocked_edges)
    observeBlocked(e.node1, e.node2);

  // Update known blocked edges in roadmap
  BOOST_FOREACH (const NodePair e, blocked_edges_) 
    roadmap.removeEdge(e.first, e.second);

  // Find plan
  PlanningResult plan = roadmap.shortestPath(req.pos, req.goal);
  
  // Return
  if (plan) {
    ROS_DEBUG_STREAM_NAMED ("det_topo_planner", "Found plan with " << plan->size() << " waypoints");
    resp.plan_found = true;
    resp.wait = false;
    resp.waypoints = *plan;
    return true;
  }
  else {
    ROS_DEBUG_STREAM_NAMED ("det_topo_planner", "Did not find a plan");
    resp.plan_found = false;
    return true;
  }
}


} // namespace topological_map


int main (int argc, char** argv)
{
  ros::init(argc, argv, "deterministic_topo_planner");
  topological_map::DeterministicTopoPlanner m;
  ROS_INFO_NAMED ("det_topo_planner", "deterministic_topo_planner node initialized");
  ros::MultiThreadedSpinner spinner(2);
  spinner.spin();
  return 0;
}
