/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file 
 * 
 * Implementation of incremental_roadmap.h
 */

#include <topological_map/incremental_roadmap.h>
#include <topological_map/utils.h>
#include <topological_map/exception.h>
#include <topological_map/point_ops.h>
#include <boost/optional.hpp>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <queue>
#include <algorithm>
#include <boost/foreach.hpp>

#define foreach BOOST_FOREACH

namespace topological_map
{

using boost::extents;
using boost::bind;
using boost::ref;
using boost::optional;
using costmap_2d::Costmap2D;
using std::min;
using std::queue;
using geometry_msgs::Point;

/************************************************************
 * Utilities
 ************************************************************/

Cell2D pointToCell (const Costmap2D& cmap, const Point& p)
{
  unsigned mx=-1;
  unsigned my=-1;
  if (cmap.worldToMap(p.x, p.y, mx, my))
    return Cell2D(mx, my);
  else 
    throw UnknownCostmapPointException(cmap, p);
}

Cell2D nodeToCell (const Roadmap& roadmap, const Costmap2D& cmap, const NodeId id) 
{
  return pointToCell(cmap, roadmap.nodePoint(id));
}

Point cellToPoint (const Costmap2D& cmap, const Cell2D& c)
{
  Point p;
  p.z=0.0;
  cmap.mapToWorld(c.r, c.c, p.x, p.y);
  ROS_ASSERT (c==pointToCell(cmap, p));
  return p;
}

vector<unsigned> getDims (const unsigned x_size, const unsigned y_size)
{
  vector<unsigned> dims(2);
  dims[0]=x_size;
  dims[1]=y_size;
  return dims;
}


/************************************************************
 * Distance propagation
 ************************************************************/

typedef map<Cell2D, unsigned> DistanceMap;
typedef DistanceMap::value_type DistanceMapEntry;

struct QueueItem
{ 
  QueueItem(const Cell2D& c, const unsigned dist) : cell(c), dist(dist) {}
  const Cell2D cell;
  const unsigned dist;
};


DistanceMap IncrementalRoadmap::cellsWithinRadius (const Cell2D& cell, const unsigned r) const
{
  queue<QueueItem> q;
  set<Cell2D> seen;
  q.push(QueueItem(cell, 0u));
  DistanceMap distances;

  ROS_DEBUG_STREAM_NAMED ("inc_roadmap_prop", "Propagating outward from " << cell << " to radius " << r);
  while (!q.empty()) {
    const QueueItem item = q.front();
    q.pop();
    if (!contains(seen, item.cell)) {
      seen.insert(item.cell);
      distances[item.cell] = item.dist;
      ROS_DEBUG_STREAM_NAMED ("inc_roadmap_prop", "Adding cell " << item.cell << " at distance " << item.dist);
      if (item.dist < r) {
        const unsigned new_dist = 1+item.dist;
        foreach (const Cell2D& neighbor, grid_graph_.neighbors(item.cell)) {
          q.push(QueueItem(neighbor, new_dist));
        }
      }
    }
  }
  return distances;
}



/************************************************************
 * Initialization
 ************************************************************/

IncrementalRoadmap::IncrementalRoadmap (const Costmap2D& cmap, double edge_distance_threshold) :
  dims_(getDims(cmap.getSizeInCellsX(), cmap.getSizeInCellsY())), resolution_(cmap.getResolution()), edge_distance_threshold_(edge_distance_threshold),
  max_waypoint_dist_(floor(edge_distance_threshold_/resolution_)), obstacle_distances_(dims_), waypoint_distances_(dims_)
{
  updateMetricMap(cmap);
  ROS_DEBUG_NAMED ("inc_roadmap", "In constructor with dims %u x %u, with resolution %.2f and edge distance threshold %.2f", dims_[0], dims_[1], resolution_, edge_distance_threshold);
}


/************************************************************
 * Waypoint addition
 ************************************************************/

// Check if a cell is already covered by a waypoint
bool IncrementalRoadmap::isCovered (const Cell2D& c) const
{
  return waypoint_distances_[c.r][c.c] <= max_waypoint_dist_;
}


Cell2D IncrementalRoadmap::bestNearbyCell (const Cell2D& c) const
{
  const unsigned max_dist = round(edge_distance_threshold_*WAYPOINT_SEARCH_MULTIPLIER/resolution_);
  unsigned best_obstacle_dist = obstacle_distances_[c.r][c.c];
  Cell2D best_cell = c;

  foreach (const DistanceMapEntry& e, cellsWithinRadius(c, max_dist)) {
    const Cell2D& cell = e.first;
    if (obstacle_distances_[cell.r][cell.c] > best_obstacle_dist) {
      best_obstacle_dist = obstacle_distances_[cell.r][cell.c];
      best_cell = cell;
    }
  }
  
  ROS_DEBUG_STREAM_NAMED ("inc_roadmap_best_cell", "Best cell near " << c << " is " << best_cell);
  
  return best_cell;
}


NodeId IncrementalRoadmap::addNode (const Cell2D& c) 
{
  ROS_DEBUG_STREAM_NAMED ("inc_roadmap_add_node", "Adding node " << c);
  set<Cell2D> new_waypoints;
  new_waypoints.insert(c);
  repropagateWaypointDistances (&waypoint_distances_, new_waypoints);
  const NodeId id = roadmap_.addNode(cellToPoint(metric_map_, c));
  connectToNearbyNodes (id);
  return id;
}


/************************************************************
 * Edge addition and pruning
 ************************************************************/


void IncrementalRoadmap::connectToNearbyNodes (const NodeId id)
{
  const Cell2D cell = nodeToCell(roadmap_, metric_map_, id);
  DistanceMap nearby_cells = cellsWithinRadius(cell, round(EDGE_DISTANCE_MULTIPLIER*edge_distance_threshold_/resolution_));
  foreach (const NodeId c2, roadmap_.allNodes()) {
    const Cell2D cell2 = nodeToCell(roadmap_, metric_map_, c2);
    if ((c2!=id) && (contains(nearby_cells, cell2))) {
      const double length = resolution_*nearby_cells[cell2];
      ROS_DEBUG_STREAM_NAMED ("inc_roadmap_add_edge", "Adding edge between nodes " << c2 << 
                              " and " << id << " with length " << length);
      roadmap_.addEdge(id, c2, length);
    }
  }
}


// Return a number that expresses how good it is to remove the edge between id1 and id2.  Higher is worse.
double IncrementalRoadmap::edgePruneCost (const NodeId id, const NodeId id2, Roadmap* roadmap_copy) const
{
  const double orig_cost = *(roadmap_copy->costBetween(id, id2));
  roadmap_copy->removeEdge(id, id2);
  optional<double> res = roadmap_copy->costBetween(id, id2);
  roadmap_copy->addEdge(id, id2, orig_cost);
  return res ? *res - orig_cost : orig_cost*10.0;
}



void IncrementalRoadmap::removeEdge (NodeId n1, NodeId n2)
{
  ROS_DEBUG_STREAM_NAMED ("inc_roadmap_remove_edge", "Removing edge between " << n1 << " and " << n2);
  roadmap_.removeEdge(n1, n2);
}


/************************************************************
 * Completely reset roadmap
 ************************************************************/

void IncrementalRoadmap::setRoadmap (const Roadmap& roadmap)
{
  roadmap_ = roadmap;
  waypoint_distances_ = computeWaypointDistances();
  ROS_DEBUG_STREAM_NAMED ("inc_roadmap_set_roadmap", "Set roadmap to one with " << roadmap_.allNodes().size() << " nodes");
}



/************************************************************
 * Metric map update
 ************************************************************/

void IncrementalRoadmap::updateMetricMap (const Costmap2D& new_map)
{
  const unsigned new_x = new_map.getSizeInCellsX();
  const unsigned new_y = new_map.getSizeInCellsY();
  const double new_res = new_map.getResolution();
  if ((new_x != dims_[0]) || (new_y != dims_[1]) || (!closeTo(new_res, resolution_))) {
    ROS_ERROR ("New %ux%u map with resolution %.8f didn't match expected params %ux%u with resolution %.8f.  Ignoring this map.",
               new_x, new_y, new_res, dims_[0], dims_[1], resolution_);
    throw TopologicalMapException("Incremental roadmap updated with incorrect-sized costmap");
  }
  else {
    ROS_DEBUG_NAMED ("inc_roadmap_update_map", "Updating metric map");
    metric_map_ = new_map;
    grid_graph_ = GridGraph(new_map);
    obstacle_distances_ = grid_graph_.obstacleDistances();
    waypoint_distances_ = computeWaypointDistances();
    ROS_DEBUG_NAMED ("inc_roadmap_update_map", "Done updating metric map");
  }
}

typedef multi_array<unsigned, 2> WaypointDistances;


WaypointDistances IncrementalRoadmap::computeWaypointDistances () const
{
  WaypointDistances distances(dims_);

  for (unsigned x=0; x<dims_[0]; ++x) {
    for (unsigned y=0; y<dims_[1]; ++y) {
      distances[x][y] = max_waypoint_dist_+1;
    }
  }

  repropagateWaypointDistances(&distances, nodeCells());
  ROS_DEBUG_NAMED ("inc_roadmap", "Done recomputing waypoint distances");
  return distances;
}


set<Cell2D> IncrementalRoadmap::nodeCells () const
{
  set<Cell2D> waypoints;
  const set<NodeId> nodes = roadmap_.allNodes();
  transform(nodes.begin(), nodes.end(), inserter(waypoints, waypoints.begin()), bind(nodeToCell, ref(roadmap_), ref(metric_map_), _1));
  return waypoints;
}


void bfsUpdate (queue<Cell2D>* q, WaypointDistances* distances, multi_array<bool, 2>* seen, const Cell2D& cell, const unsigned new_cost)
{
  const unsigned r = cell.r;
  const unsigned c = cell.c;
  (*seen)[r][c] = true;
  if ((*distances)[r][c] > new_cost) {
    (*distances)[r][c] = new_cost;
    q->push(cell);
  }
}



void IncrementalRoadmap::repropagateWaypointDistances(WaypointDistances* distances, const set<Cell2D>& cells) const
{
  foreach (const Cell2D& waypoint, cells) {
    foreach (DistanceMapEntry e, cellsWithinRadius(waypoint, max_waypoint_dist_)) {
      const Cell2D& cell = e.first;
      const unsigned dist = e.second;
      if (dist < (*distances)[cell.r][cell.c])
        (*distances)[cell.r][cell.c] = dist;
    }
  }
}



/************************************************************
 * Accessors
 ************************************************************/

unsigned IncrementalRoadmap::getXSize () const
{
  return dims_[0];
}

unsigned IncrementalRoadmap::getYSize () const
{
  return dims_[1];
}

double IncrementalRoadmap::getResolution () const
{
  return resolution_;
}

const Roadmap& IncrementalRoadmap::roadmapSnapshot () const
{
  return roadmap_;  
}

bool IncrementalRoadmap::knownFree (const Cell2D& c) const
{
  return grid_graph_.containsCell(c);
}





} // topological_map
