/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2009, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

#include <topological_map/roadmap.h>
#include <topological_map/MoveBaseTopoAction.h>
#include <topological_map/point_ops.h>
#include <topological_map/utils.h>
#include <topological_map/roadmap_msg.h>
#include <topological_map/GetTopologicalPlan.h>
#include <ros/ros.h>
#include <tf/transform_listener.h>
#include <actionlib/server/simple_action_server.h>
#include <actionlib/client/simple_action_client.h>
#include <actionlib/client/simple_client_goal_state.h>
#include <visualization_msgs/Marker.h>
#include <move_base_msgs/MoveBaseAction.h>
#include <nav_msgs/GetPlan.h>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <boost/optional.hpp>
#include <boost/thread.hpp>
#include <boost/foreach.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/functional.hpp>
#include <ostream>
#include <string>

/****************************************
 * FILEWIDE DEFINE
 ****************************************/
#define foreach BOOST_FOREACH



namespace topological_map {

using boost::bind;
using boost::ref;
using boost::optional;
using ros::Time;
using ros::Rate;
using geometry_msgs::Pose;
using geometry_msgs::PoseStamped;
using std::ostream;
using std::string;
using geometry_msgs::Point;
using std::pair;
using visualization_msgs::Marker;

typedef actionlib::SimpleClientGoalState GoalState;
typedef actionlib::SimpleActionServer<MoveBaseTopoAction> ActionServer;
typedef actionlib::SimpleActionClient<move_base_msgs::MoveBaseAction> ActionClient;
typedef vector<NodeId> NodeVec;
typedef boost::shared_mutex Mutex;
typedef boost::shared_lock<Mutex> ReadLock;
typedef boost::unique_lock<Mutex> WriteLock;

const string DEFAULT_FIXED_FRAME="/map";
const string DEFAULT_BASE_FRAME="base_link";
const double EDGE_DISTANCE_MULTIPLIER=3.0; // If an edge length is this factor greater than the Euclidean distance, it's considered "blocked"
const double WAIT = -1; // TODO this needs to be handled better
const double GOAL_CONNECT_DISTANCE=2.0;
const double DEFAULT_EXIT_RADIUS=2.0;
const double DEFAULT_GOAL_DISTANCE_THRESHOLD=0.3;
const double TOP_LEVEL_LOOP_RATE=1.0;

/************************************************************
 * Node class
 ************************************************************/

class MoveBaseTopo
{
public:
  /// Constructor
  MoveBaseTopo();

private:

  /****************************************
   * Callbacks
   ****************************************/
  
  /// The action callback
  void executeCB (const MoveBaseTopoGoal::ConstPtr& goal);

  /// Simple (non-action) goal callback
  void simpleGoalCB (const PoseStamped::ConstPtr& goal);
  
  /// Roadmap callback
  void roadmapCB (const RoadmapRos::ConstPtr& roadmap);

  /****************************************
   * Const ops
   ****************************************/

  /// Return the id of the node that's closest in shortest-path distance to \a p
  /// Will assert if the roadmap is empty
  NodeId getClosestNode (const Point& p) const;

  /// Compute shortest path distance between two points using navigation planner
  /// \todo Currently stubbed out with Euclidean distance
  double shortestPathDistance (const Point& p1, const Point& p2) const;

  /// Returns true iff the given edge is blocked
  /// Blocked means the shortestPathDistance is much higher than the Euclidean distance
  bool isBlocked (NodeId n1, NodeId n2) const;

  /// Get the set of incident nodes to \a n the edges to which are blocked 
  set<NodeId> blockedEdges (NodeId n) const;

  /// Plan from \a n to \a goal assuming the edges to \a blocked are currently observed blocked
  optional<NodeVec> getPlan (const NodeId n, const set<NodeId>& blocked, const NodeId goal);
  
  /// Return current position (threadsafely)
  Point getCurrentPos () const;

  /// Return last node on path that's within the exit radius
  NodeId getExitNode (const NodeVec& path, const Point& pos) const;

  /****************************************
   * Updating ops
   ****************************************/

  /// Update the robot's current position
  void updatePosition (const ros::WallTimerEvent& e);

  /// Connect a newly added goal node to nearby nodes
  void connectToNearbyNodes (const NodeId n);

  /// Visualize the roadmap
  void visualizeRoadmap (const ros::WallTimerEvent& e);

  /// Publish a single marker
  void publishMarker (Marker* marker);



  /****************************************
   * Params
   ****************************************/

  const double exit_radius_;
  const double goal_distance_threshold_;

  /****************************************
   * Associated objects
   ****************************************/
  
  ros::NodeHandle nh_;
  ActionServer mb_topo_server_;
  ActionClient mb_client_;
  ros::Subscriber simple_goal_sub_, roadmap_sub_;
  ros::Publisher action_goal_pub_, vis_pub_;
  mutable ros::ServiceClient navfn_client_, topo_plan_client_;
  tf::TransformListener tf_;
  mutable Mutex roadmap_mutex_, pose_mutex_;
  ros::WallTimer visualization_timer_, position_update_timer_;

  /****************************************
   * State
   ****************************************/

  // The most recent roadmap published by the roadmap server
  Roadmap roadmap_;

  // The robot's position in the map frame
  Point current_pos_;

  // Will hold the current path (once it's found)
  optional<NodeVec> path_;

};


/************************************************************
 * Initialization
 ************************************************************/

double getPrivateParam(const string& name, const double default_val)
{
  double t;
  ros::NodeHandle nh("~");
  nh.param(name, t, default_val);
  return t;
}

MoveBaseTopo::MoveBaseTopo () : 
  exit_radius_(getPrivateParam("exit_radius", DEFAULT_EXIT_RADIUS)),
  goal_distance_threshold_(getPrivateParam("goal_distance_threshold", DEFAULT_GOAL_DISTANCE_THRESHOLD)),
  mb_topo_server_("move_base_topo", bind(&MoveBaseTopo::executeCB, this, _1)),
  mb_client_("move_base", true),
  simple_goal_sub_(nh_.subscribe("move_base_topo/simple_goal", 10, &MoveBaseTopo::simpleGoalCB, this)),
  roadmap_sub_(nh_.subscribe("roadmap", 10, &MoveBaseTopo::roadmapCB, this)),
  action_goal_pub_(nh_.advertise<MoveBaseTopoActionGoal>("move_base_topo/goal", 10)),
  vis_pub_(nh_.advertise<Marker>("visualization_marker", 10)),
  navfn_client_(nh_.serviceClient<nav_msgs::GetPlan>("move_base_node/NavfnROS/make_plan")),
  topo_plan_client_(nh_.serviceClient<GetTopologicalPlan>("get_topological_plan")),
  visualization_timer_(nh_.createWallTimer(ros::WallDuration(3.0), &MoveBaseTopo::visualizeRoadmap, this)),
  position_update_timer_(nh_.createWallTimer(ros::WallDuration(0.5), &MoveBaseTopo::updatePosition, this))
{
}
                                

/************************************************************
 * State updates
 ************************************************************/

// Just set the roadmap from the callback
// If the roadmap's locked by the executive, don't set it
void MoveBaseTopo::roadmapCB (const RoadmapRos::ConstPtr& roadmap)
{
  
  WriteLock l(roadmap_mutex_, boost::try_to_lock_t());
  if (l.owns_lock()) {
    const unsigned num_nodes_old = roadmap_.allNodes().size();
    roadmap_ = message::roadmapFromRos(*roadmap);
    const unsigned num_nodes = roadmap_.allNodes().size();
    if ((num_nodes_old==0) && (num_nodes>0))
      ROS_DEBUG_NAMED ("move_base_topo", "Received first roadmap, with %u nodes", num_nodes);
  }
}


// Set the position using tf
void MoveBaseTopo::updatePosition (const ros::WallTimerEvent& e)
{
  ros::Time now = ros::Time::now();
  if (tf_.waitForTransform(DEFAULT_FIXED_FRAME, DEFAULT_BASE_FRAME, now, ros::Duration(0.1))) {
    tf::Stamped<tf::Pose> base_pose, fixed_pose;
    base_pose.setIdentity();
    base_pose.frame_id_ = DEFAULT_BASE_FRAME;
    base_pose.stamp_ = now;
    
    // TODO: This can technically throw an exception
    tf_.transformPose(DEFAULT_FIXED_FRAME, base_pose, fixed_pose);

    PoseStamped pose;
    tf::poseStampedTFToMsg(fixed_pose, pose);
    WriteLock lock(pose_mutex_);
    current_pos_ = pose.pose.position;    
  }
  else {
    ROS_INFO_STREAM_NAMED ("move_base_topo", "Skipping pose update at " << now <<
                           " because didn't receive a transform by " << ros::Time::now());
  }
}

Point MoveBaseTopo::getCurrentPos () const
{
  ReadLock lock(pose_mutex_);
  return current_pos_;
}
                
/************************************************************
 * Geometry 
 ************************************************************/

// Helper for outputting points
// TODO: this will no longer be necessary post ros 1.1, where operator<< will be defined
ostream& operator<< (log4cxx::helpers::CharMessageBuffer& str, const Point& p)
{
  str << "[" << p.x << ", " << p.y << ", " << p.z << "]";
  return str;
}



// Comparison function when sorting nodes according to distance from a point
bool closerToPoint (const Roadmap& roadmap, const Point& p, const NodeId n1, const NodeId n2)
{
  return euclideanDistance(p, roadmap.nodePoint(n1)) < euclideanDistance(p, roadmap.nodePoint(n2));
}
                

// Calls out to move_base's navigation planner to compute distances
double MoveBaseTopo::shortestPathDistance (const Point& p1, const Point& p2) const
{
  nav_msgs::GetPlan srv;
  geometry_msgs::Quaternion identity;
  identity.w = 1.0;
  identity.x = identity.y = identity.z = 0.0;
  srv.request.start.header.frame_id = DEFAULT_FIXED_FRAME;
  srv.request.start.header.stamp = ros::Time::now();
  srv.request.goal.header = srv.request.start.header;
  srv.request.tolerance = goal_distance_threshold_/2;
  srv.request.start.pose.position = p1;
  srv.request.goal.pose.position = p2;
  srv.request.start.pose.orientation = srv.request.goal.pose.orientation = identity;
  ROS_DEBUG_STREAM_NAMED ("move_base_topo_distance", "Calling navigation service with start " 
                          << p1 << " and goal " << p2);
  if (navfn_client_.call(srv)) {
    const unsigned num_waypoints = srv.response.plan.poses.size();
    ROS_DEBUG_STREAM_NAMED ("move_base_topo_distance", "Returned path with " 
                            << num_waypoints << " waypoints");
    if (num_waypoints<2) {
      return euclideanDistance(p1, p2);
    }
    else {
      double length=0.0;
      for (unsigned i=0; i<num_waypoints-1; i++)
        length += euclideanDistance(srv.response.plan.poses[i].pose.position, 
                                    srv.response.plan.poses[i+1].pose.position);
      return length;
    }
  }
  else {
    ROS_WARN_STREAM ("Service call to local navigation failed given " << p1 <<
                     " and " << p2 <<"; falling back to Euclidean distance.");
    return euclideanDistance (p1, p2);
  }
}

Edge makeBlockedEdge (const NodeId n1, const NodeId n2)
{
  Edge e;
  e.node1 = n1;
  e.node2 = n2;
  return e;
}



vector<NodeId> nodesInDistanceOrder (const Roadmap& roadmap, const Point& p)
{
  const set<NodeId> nodes = roadmap.allNodes();
  vector<NodeId> nodes_sorted(nodes.begin(), nodes.end());
  sort(nodes_sorted.begin(), nodes_sorted.end(), bind(closerToPoint, ref(roadmap), ref(p), _1, _2));
  return nodes_sorted;
}

// Find the closest node to p, in shortest-path distance
NodeId MoveBaseTopo::getClosestNode (const Point& p) const
{
  vector<NodeId> nodes_sorted = nodesInDistanceOrder(roadmap_, p);

  // Now loop in order of increasing Euclidean distance
  optional<NodeId> best_node;
  double best_distance=-42.42; // valid iff best_node is set
  foreach (const NodeId n, nodes_sorted) {

    // If this is the first node, or if this node is the closest so far, set best_node 
    // and best_distance. Use fact that Euclidean distance lower bounds shortest path 
    // distance to only call the latter when necessary
    if (!best_node || euclideanDistance(p, roadmap_.nodePoint(n)) < best_distance) {
      const double current_distance = shortestPathDistance(p, roadmap_.nodePoint(n));
      if (!best_node || current_distance<best_distance) {
        best_node = n;
        best_distance = current_distance;
      }
    }
  }

  ROS_ASSERT (best_node);
  return *best_node;
}

bool withinRadius (const double r, const Roadmap& roadmap, const Point& p, NodeId n)
{
  const Point p2 = roadmap.nodePoint(n);
  const double dist = euclideanDistance(p, p2);
  ROS_DEBUG_STREAM_NAMED ("move_base_topo_exit_pt", "Distance from " << p << " to "
                          << p2 << " is " << dist);
  return dist < r;
}


NodeId MoveBaseTopo::getExitNode (const NodeVec& path, const Point& pos) const
{
  NodeVec::const_reverse_iterator iter = 
    find_if(path.rbegin(), path.rend(), bind(withinRadius, exit_radius_, ref(roadmap_), ref(pos), _1));
  if (iter==path.rend()) {
    ROS_DEBUG_STREAM_NAMED ("move_base_node", "No waypoint was close to " << pos <<
                            "; using the first one.");
    iter = path.rend()-1;
  }
  return iter==path.rbegin() ? *iter : *(iter-1);
}
    



/************************************************************
 * Blocked edges
 ************************************************************/

bool MoveBaseTopo::isBlocked (const NodeId n1, const NodeId n2) const
{
  const Point p1 = roadmap_.nodePoint(n1);
  const Point p2 = roadmap_.nodePoint(n2);
  return shortestPathDistance(p1, p2) > EDGE_DISTANCE_MULTIPLIER*euclideanDistance(p1, p2);
}

set<NodeId> MoveBaseTopo::blockedEdges (NodeId n) const
{
  const set<NodeId> neighbors = roadmap_.neighbors(n);
  set<NodeId> blocked;
  remove_copy_if (neighbors.begin(), neighbors.end(), inserter(blocked, blocked.begin()),
                  !bind(&MoveBaseTopo::isBlocked, this, n, _1));
  return blocked;
}

/************************************************************
 * Planning
 ************************************************************/

// RAII object for temporarily removing an edge from the graph
// TODO get rid of TemporaryBlockedEdge, which is no longer used
class TemporaryBlockedEdge
{
public:
  TemporaryBlockedEdge (const NodeId n1, const NodeId n2, Roadmap* roadmap) :
    n1(n1), n2(n2), roadmap(roadmap), cost(roadmap->getCost(n1, n2))
  { 
    roadmap->removeEdge(n1, n2);
  }

  ~TemporaryBlockedEdge () { roadmap->addEdge(n1, n2, cost); }

private:
  const NodeId n1, n2;
  Roadmap* roadmap;
  const double cost;
};

typedef boost::shared_ptr<TemporaryBlockedEdge> TempBlockEdgePtr;


optional<NodeVec> MoveBaseTopo::getPlan (const NodeId n, const set<NodeId>& blocked, const NodeId goal)
{
  // Populate the service request
  GetTopologicalPlan srv;
  srv.request.graph = message::roadmapToRos(roadmap_);
  srv.request.pos = n;
  srv.request.goal = goal;
  transform (blocked.begin(), blocked.end(), back_inserter(srv.request.blocked_edges),
             bind(makeBlockedEdge, n, _1));


  // Call the service and return an answer if one was found
  if (topo_plan_client_.call(srv)) {
    if (srv.response.plan_found) {
      if (srv.response.wait) {
        ROS_WARN ("Returning a wait action from topological planner not currently supported");
      }
      else {
        NodeVec plan(srv.response.waypoints.begin(), srv.response.waypoints.end());
        return optional<NodeVec>(plan);
      }
    }
    else {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_plan", "Plan not found between " << n << " and " << goal);
    }
  }
  else {
    ROS_WARN_NAMED ("move_base_topo_plan", "Service call to topological planner failed.");
  }

  // If service call failed or plan was not found, return nothing
  return optional<NodeVec> ();
}


  

/************************************************************
 * Action server
 ************************************************************/

void MoveBaseTopo::connectToNearbyNodes (const NodeId n)
{
  const Point p = roadmap_.nodePoint(n);
  vector<NodeId> nodes = nodesInDistanceOrder(roadmap_, p);
  foreach (const NodeId n2, nodes) {
    const Point p2 = roadmap_.nodePoint(n2);
    if (euclideanDistance(p, p2) > GOAL_CONNECT_DISTANCE)
      break;
    else {
      const double distance = shortestPathDistance(p, p2);
      if ((n2!=n) && (distance < GOAL_CONNECT_DISTANCE))
        roadmap_.addEdge(n, n2, distance);
    }
  }
}


void MoveBaseTopo::executeCB (const MoveBaseTopoGoal::ConstPtr& goal)
{
  ROS_DEBUG_STREAM_NAMED ("move_base_topo", "Received goal " << goal->target_pose.pose.position);

  NodeId goal_node=-1;
  boost::upgrade_lock<Mutex> l(roadmap_mutex_);
  {
    boost::upgrade_to_unique_lock<Mutex> l2(l);
    goal_node = roadmap_.addNode(goal->target_pose.pose.position);
    connectToNearbyNodes(goal_node);
  }
  
  // Holds the set of blocked edges from the current closest node
  set<NodeId> blocked;


  /****************************************
   * The execution loop for this goal
   ****************************************/

  while (ros::ok()) {
    const Point pos = getCurrentPos();
    const NodeId closest = getClosestNode(pos);
    ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "Current pos is " << pos <<
                            ", closest node is " << closest << " at " << roadmap_.nodePoint(closest));
    if (path_) {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "Current path is " << toString(*path_));
    }
    else {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "No path found yet");
    }
    ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "Blocked nodes are " << toString(blocked));
    ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "Client status is " << mb_client_.getState().toString());

    /****************************************
     * Case 1: we're at the goal
     ****************************************/
    if (euclideanDistance(pos, goal->target_pose.pose.position) < goal_distance_threshold_) {
      ROS_DEBUG_NAMED ("move_base_topo_exec", "Reached %.2f distance from goal; success!",
                       euclideanDistance(pos, goal->target_pose.pose.position));
      MoveBaseTopoResult result;
      mb_topo_server_.setSucceeded(result);
      break;
    }

    /****************************************
     * Case 2: we're preempted
     ****************************************/
    else if (mb_topo_server_.isPreemptRequested()) {
      ROS_DEBUG_NAMED ("move_base_topo_exec", "Action preempted");
      mb_topo_server_.setPreempted();
      break;
    }

    /****************************************
     * Case 3: if we don't have a path
     * that begins at the closest node,
     * recompute blocked edges and replan
     ****************************************/
    else if (!path_ || *(path_->begin()) != closest) {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", 
                              "Path nonexistent or doesn't match nearest node");
      blocked = blockedEdges(closest);
      path_ = getPlan(closest, blocked, goal_node);
    }


    /****************************************
     * Case 4: if first action is to wait,
     * do so, then spin around, recompute
     * blocked edges, and replan
     ****************************************/
    else if (*(path_->begin()) == WAIT) {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec",
                              "First action is WAIT (this shouldn't happen yet)");
      // spinAround();
      blocked = blockedEdges(closest);
      path_ = getPlan(closest, blocked, goal_node);
    }


    /****************************************
     * Case 5: if current edge is blocked,
     * update state and replan
     ****************************************/
    else if ((path_->size()>1) && isBlocked((*path_)[0], (*path_)[1])) {
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "Edge from " << (*path_)[0] << " to " <<
                              (*path_)[1] << " is blocked; replanning");
      blocked.insert((*path_)[1]);
      path_ = getPlan(closest, blocked, goal_node);
    }

    /****************************************
     * Case 6: send the current waypoint
     * as a goal to move base
     ****************************************/
    else {
      move_base_msgs::MoveBaseGoal mb_goal;
      const NodeId exit_node = getExitNode(*path_, pos);
      mb_goal.target_pose.pose.position = roadmap_.nodePoint(exit_node);
      mb_goal.target_pose.pose.orientation.w = 1.0;
      mb_goal.target_pose.header.frame_id = DEFAULT_FIXED_FRAME;
      mb_goal.target_pose.header.stamp = ros::Time::now();
      mb_client_.sendGoal(mb_goal);
      ROS_DEBUG_STREAM_NAMED ("move_base_topo_exec", "All ok... sending goal " << 
                              mb_goal.target_pose.pose.position);
    }
     
    ros::Duration(1/TOP_LEVEL_LOOP_RATE).sleep();
  }
  
  { 
    boost::upgrade_to_unique_lock<Mutex> l2(l);
    roadmap_.removeNode(goal_node);
  }
}
      


// Republish simple pose stamped goals to the action server
void MoveBaseTopo::simpleGoalCB (const PoseStamped::ConstPtr& goal)
{
  ROS_DEBUG_STREAM_NAMED ("move_base_topo", "Republishing simple goal to the action server: " << goal->pose.position);
  MoveBaseTopoActionGoal action_goal;
  action_goal.goal.target_pose = *goal;
  action_goal_pub_.publish(action_goal);
  ROS_DEBUG_STREAM_NAMED ("move_base_topo", "Done simple goal ");// << goal->pose.position);

}





/************************************************************
 * Visualization
 ************************************************************/

bool onPath (const NodeVec& path, const NodeId id, const NodeId id2)
{
  typedef NodeVec::const_iterator NVIter;

  NVIter pos = find(path.begin(), path.end(), id);
  NVIter pos2 = find(path.begin(), path.end(), id2);
  
  return 
    !(pos==path.end() || pos2==path.end()) &&
    ((pos!=path.begin()) && (pos-1==pos2)) ||
    ((pos2!=path.begin()) && (pos2-1==pos));
}

void MoveBaseTopo::publishMarker (Marker* marker) 
{
  if (marker->points.size() == 0) {
    marker->action = visualization_msgs::Marker::DELETE;
  }
  marker->header.stamp = ros::Time::now();
  vis_pub_.publish(*marker);
}

void MoveBaseTopo::visualizeRoadmap (const ros::WallTimerEvent& e) 
{
  ROS_DEBUG_NAMED ("move_base_topo_vis", "Visualization waiting for lock");
  ReadLock lock(roadmap_mutex_);

  // Scope in which lock on waypoint_graph_ is held
  Marker nodes, edges;

  ROS_DEBUG_NAMED ("move_base_topo_vis", "Visualizing in move_base_topo");

  nodes.header.frame_id = "/map";
  nodes.id = 0;
  nodes.ns = "roadmap_node";
  nodes.type = visualization_msgs::Marker::SPHERE_LIST;
  nodes.action = visualization_msgs::Marker::ADD;
  nodes.scale.x = 0.2;
  nodes.scale.y = 0.2;
  nodes.scale.z = 0.2;
  nodes.color.r = 1.0;
  nodes.color.a = 1.0;

  edges.header.frame_id = "/map";
  edges.id = 1;
  edges.ns = "roadmap_node";
  edges.type = visualization_msgs::Marker::LINE_LIST;
  edges.action = visualization_msgs::Marker::ADD;
  edges.scale.x = 0.1;
  edges.color.b = 1.0;
  edges.color.g = 0.5;
  edges.color.a = 1.0;

  Marker active_edges = edges;
  active_edges.id = 2;
  active_edges.scale.x = 0.2;
  active_edges.color.b = .1;
  active_edges.color.r = 0.5;



  foreach (const NodeId id, roadmap_.allNodes()) {
    const Point p = roadmap_.nodePoint(id);
    nodes.points.push_back(p);

    foreach (NodeId id2, roadmap_.neighbors(id)) {
      if (id<id2) {
        const Point q = roadmap_.nodePoint(id2);
        if (path_ && onPath(*path_, id, id2)) {
          active_edges.points.push_back(p);
          active_edges.points.push_back(q);
        }
        else {
          edges.points.push_back(p);
          edges.points.push_back(q);
        }
      }
    }
  }

  publishMarker(&nodes);
  publishMarker(&edges);
  publishMarker(&active_edges);
}




} // namespace topological_map


int main (int argc, char** argv)
{
  ros::init(argc, argv, "move_base_topo");
  topological_map::MoveBaseTopo m;
  ROS_INFO ("move_base_topo node initialized");
  ros::MultiThreadedSpinner spinner(4);
  spinner.spin();
  return 0;
}
