/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file
 *
 * Implements internally used Roadmap data structure
 *
 * \author Bhaskara Marthi
 */

#include <topological_map/roadmap.h>
#include <topological_map/point_ops.h>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <algorithm>
#include <ros/console.h>
#include <ros/assert.h>
#include <topological_map/exception.h>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <boost/foreach.hpp>

#define foreach BOOST_FOREACH

namespace topological_map
{

using boost::tie;
using boost::bind;
using boost::ref;
using boost::optional;
using std::map;
using std::endl;
using geometry_msgs::Point;

typedef map<RoadmapVertex, double> DistanceMap;
typedef map<RoadmapVertex, RoadmapVertex> PredecessorMap;
typedef boost::associative_property_map<DistanceMap> DistancePmap;
typedef boost::associative_property_map<PredecessorMap> PredecessorPmap;


/************************************************************
 * Basic ops
 ************************************************************/

NodeId Roadmap::addNode (const Point& p)
{
  while (idExists(next_id_))
    next_id_++;
  addNode(p, next_id_);
  return next_id_;
}

void Roadmap::addNode (const Point& p, const NodeId id)
{
  if (id_vertex_map_.find(id)!=id_vertex_map_.end())
    throw DuplicateNodeIdException(id);
  id_vertex_map_[id]=add_vertex(NodeInfo(id, p), graph_);
  ROS_DEBUG_STREAM_NAMED ("roadmap", "Added node " << id << " at " << p);
}

void Roadmap::addEdge (const NodeId i, const NodeId j, double cost)
{
  if (edge(idVertex(i), idVertex(j), graph_).second)
    throw DuplicateEdgeException(i, j);
  const RoadmapEdge e = ensureEdge(idVertex(i),idVertex(j));
  graph_[e].cost = cost;
  ROS_DEBUG_STREAM_NAMED ("roadmap", "Set cost between nodes " << i << " and " << j << " to " << cost);
}

double Roadmap::getCost (const NodeId i, const NodeId j) const
{
  pair<RoadmapEdge, bool> res = edge(idVertex(i), idVertex(j), graph_);
  if (!res.second)
    throw NonexistentEdgeException (i, j);
  return graph_[res.first].cost;
}
    


void Roadmap::removeNode (const NodeId i)
{
  const RoadmapVertex v = idVertex(i);
  clear_vertex(v, graph_);
  remove_vertex(v, graph_);
  id_vertex_map_.erase(id_vertex_map_.find(i));
  ROS_DEBUG_STREAM_NAMED ("roadmap", "Removed node " << i);
}

void Roadmap::removeEdge (const NodeId i, const NodeId j)
{
  const RoadmapVertex vi = idVertex(i);
  const RoadmapVertex vj = idVertex(j);
  if (!edge(vi, vj, graph_).second)
    throw NonexistentEdgeException(i, j);
  remove_edge (vi, vj, graph_);
}


struct CorrespondsToId 
{
  CorrespondsToId (const RoadmapImpl& graph, const NodeId id) : graph(graph), id(id) {}
  bool operator() (const RoadmapVertex& v) { return graph[v].id==id; }
  const RoadmapImpl& graph;
  const NodeId id;
};

Point Roadmap::nodePoint (const NodeId id) const
{
  VertexIterator iter, end;
  tie (iter, end) = vertices(graph_);
  VertexIterator pos = find_if(iter, end, CorrespondsToId(graph_, id));
  if (pos==end) {
    throw UnknownNodeIdException(id);
  }
  return graph_[*pos].point;
}

typedef pair<const NodeId, RoadmapVertex> Entry;
NodeId getKey (Entry e)
{
  return e.first;
}


set<NodeId> Roadmap::allNodes () const
{
  set<NodeId> nodes;
  transform (id_vertex_map_.begin(), id_vertex_map_.end(), inserter(nodes, nodes.begin()), getKey);
  return nodes;
}

NodeId vertexId (const RoadmapImpl& graph, const RoadmapVertex& v)
{
  return graph[v].id;
}

set<NodeId> Roadmap::neighbors (const NodeId id) const
{
  RoadmapAdjacencyIterator adj_iter, adj_end;
  tie(adj_iter, adj_end) = adjacent_vertices(idVertex(id), graph_);
  set<NodeId> neighbors;
  transform(adj_iter, adj_end, inserter(neighbors, neighbors.begin()), bind(vertexId, ref(graph_), _1));
  return neighbors;
}


// Get nodes within Euclidean radius of a point
// Just looks at all nodes right now
set<NodeId> Roadmap::nodesWithinRadius (geometry_msgs::Point p, double radius) const
{
  set<NodeId> s;
  foreach (const NodeId n, allNodes()) {
    if (euclideanDistance(nodePoint(n), p) < radius) 
      s.insert(n);
  }
  return s;
}

/************************************************************
 * Shortest paths
 ************************************************************/


struct DijkstraVisitor : public boost::default_dijkstra_visitor
{
  DijkstraVisitor () {}
  void discover_vertex(const RoadmapVertex& v, const RoadmapImpl& graph) const
  {
    ROS_DEBUG_STREAM_NAMED("dijkstra", "Discovering vertex " << graph[v].id << " at point " << graph[v].point);
  }
};


vector<NodeCost> Roadmap::nodeCosts (const NodeId i, const NodeId j)
{
  ROS_DEBUG_STREAM_NAMED ("roadmap_shortest_path", "Looking for node costs between " << i << " and " << j);
  const RoadmapVertex v=idVertex(i);
  const RoadmapVertex w=idVertex(j);
  resetIndices();
  DistanceMap distances;
  vector<NodeCost> costs;
  PredecessorMap predecessors; // We shouldn't actually need this, but adding it removes a warning about boost/dijkstra_shortest_paths.hpp...
  
  dijkstra_shortest_paths(graph_, w, weight_map(get(&EdgeInfo::cost, graph_)).
                          vertex_index_map(get(&NodeInfo::index, graph_)).
                          distance_map(DistancePmap(distances)).visitor(DijkstraVisitor()).
                          predecessor_map(PredecessorPmap(predecessors)));

  RoadmapAdjacencyIterator adj_iter, adj_end;
  for (tie(adj_iter, adj_end)=adjacent_vertices(v, graph_); adj_iter!=adj_end; ++adj_iter) {
    DistanceMap::iterator pos = distances.find(*adj_iter);
    if (pos!=distances.end()) {
      NodeId id = graph_[*adj_iter].id;
      double edge_cost = graph_[edge(*adj_iter, v, graph_).first].cost;
      ROS_DEBUG_STREAM_NAMED ("roadmap_shortest_path", " Node " << id << " has cost " << edge_cost+pos->second << "=" << edge_cost << "+" << pos->second);
      shortestPath(id,j);
      costs.push_back(NodeCost(id, edge_cost+pos->second));
    }
  }
  return costs;
}


optional<double> Roadmap::costBetween (const NodeId i, const NodeId j)
{
  ROS_DEBUG_STREAM_NAMED ("roadmap_shortest_path", "Looking for shortest path between roadmap nodes " << i << " and " << j);
  
  const RoadmapVertex v = idVertex(i);
  const RoadmapVertex w = idVertex(j);

  resetIndices();
  DistanceMap distances;
  PredecessorMap predecessors; // We shouldn't actually need this, but adding it removes a warning about boost/dijkstra_shortest_paths.hpp...

  dijkstra_shortest_paths(graph_, v, weight_map(get(&EdgeInfo::cost, graph_)).
                          vertex_index_map(get(&NodeInfo::index, graph_)).
                          distance_map(DistancePmap(distances)).visitor(DijkstraVisitor()).
                          predecessor_map(PredecessorPmap(predecessors)));

  optional<double> cost;
  if (distances.find(w)==distances.end()) {
    ROS_DEBUG_NAMED ("roadmap_shortest_path", "Path not found");
  }
  else {
    ROS_DEBUG_STREAM_NAMED ("roadmap_shortest_path", "Path found with length " << distances[w]);
    cost = distances[w];
  }
  return cost;
}


struct TransformVertexToId
{
  TransformVertexToId(const RoadmapImpl& graph) : graph(graph) {}
  NodeId operator() (const RoadmapVertex& v) { return graph[v].id; }
  const RoadmapImpl& graph;
};

                         
optional<vector<NodeId> > Roadmap::shortestPath (const NodeId i, const NodeId j)
{
  const RoadmapVertex v = idVertex(i);
  const RoadmapVertex w = idVertex(j);

  resetIndices();
  PredecessorMap predecessors;

  dijkstra_shortest_paths(graph_, v, weight_map(get(&EdgeInfo::cost, graph_)).
                          vertex_index_map(get(&NodeInfo::index, graph_)).
                          predecessor_map(PredecessorPmap(predecessors)).
                          visitor(DijkstraVisitor()));
    


  // Extract path from predecessor map
  vector<RoadmapVertex> reverse_path(1,w);
  RoadmapVertex current = w;

  while (current!=predecessors[current]) {
    ROS_DEBUG_NAMED("roadmap_dijkstra", "  Predecessor of %d is %d with cost %f", graph_[current].id, 
                    graph_[predecessors[current]].id, graph_[edge(current, predecessors[current], graph_).first].cost);
    reverse_path.push_back(current=predecessors[current]);
  }
  optional<vector<NodeId> > path;

  if (current==v){
    // If a path exists, put it in source-to-goal order, replace vertex descriptors with ids, and return
    path = vector<NodeId> (reverse_path.size());
    transform(reverse_path.rbegin(), reverse_path.rend(), path->begin(), TransformVertexToId(graph_));
  }
  return path; // Will return the path if it exists, or an unset value otherwise
}


/************************************************************
 * i/o
 ************************************************************/


// Assignment operator
Roadmap& Roadmap::operator=(const Roadmap& r) {
  if (this!=&r) {

    // Remove existing nodes
    foreach (const NodeId id, allNodes())    
      removeNode(id);

    // Add the new ones
    foreach (const NodeId id, r.allNodes())
      addNode(r.nodePoint(id), id);

    // Add the new edges
    foreach (const NodeId id, r.allNodes()) {
      foreach (const NodeId n, r.neighbors(id)) {
        if (id < n)
          addEdge(id, n, r.getCost(id, n));
      }
    }

  }
  return *this;
}


// Copy constructor
Roadmap::Roadmap (const Roadmap& r) : next_id_(1)
{
  // Add nodes
  foreach (const NodeId id, r.allNodes())
    addNode(r.nodePoint(id), id);

  // Add edges
  foreach (const NodeId id, r.allNodes()) {
    foreach (const NodeId n, r.neighbors(id)) {
      if (id < n)
        addEdge(id, n, r.getCost(id, n));
    }
  }
}
  



/************************************************************
 * Temporary node
 ************************************************************/

TemporaryRoadmapNode::TemporaryRoadmapNode (Roadmap* roadmap, const Point& p):
  roadmap_(roadmap), id_(roadmap->addNode(p))
{
  ROS_DEBUG_STREAM_NAMED ("temp_node", "Created temporary roadmap node with id " << id_);
}

NodeId TemporaryRoadmapNode::getId () const
{
  return id_;
}

TemporaryRoadmapNode::~TemporaryRoadmapNode ()
{
  ROS_DEBUG_STREAM_NAMED ("temp_node", "Deleted temporary roadmap node with id " << id_);
  roadmap_->removeNode(id_);
}


/************************************************************
 * internal
 ************************************************************/

RoadmapEdge Roadmap::ensureEdge(const RoadmapVertex v, const RoadmapVertex w)
{
  pair<RoadmapEdge, bool> p = edge(v,w,graph_);
  if (!(p.second)) {
    add_edge(v,w,graph_);
    p = edge(v,w,graph_);
    ROS_ASSERT(p.second);
  }
  return p.first;
}

bool Roadmap::idExists (const NodeId i) const
{
  return id_vertex_map_.find(i)!=id_vertex_map_.end();
}

RoadmapVertex Roadmap::idVertex(const NodeId i) const
{
  if (!idExists(i)) {
    throw UnknownNodeIdException(i);
  }
  else {
    return id_vertex_map_.find(i)->second;
  }
}

void Roadmap::resetIndices()
{
  VertexIterator iter, end;
  unsigned ind=0;
  for (tie(iter,end) = vertices(graph_); iter!=end; ++iter) {
    graph_[*iter].index=ind++;
  }
}











} // namespace
