/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file 
 * 
 * Implementation of roadmap_msg.h
 */


#include <topological_map/roadmap_msg.h>
#include <topological_map/exception.h>
#include <topological_map/point_ops.h>
#include <boost/foreach.hpp>
#include <rosrecord/Recorder.h>
#include <rosrecord/Player.h>
#include <ros/time.h>

#define foreach BOOST_FOREACH

namespace topological_map
{
using std::string;
using geometry_msgs::Point;

const string DUMMY_TOPIC("dummy");


// Reading from a file means reading from the corresponding bag
// Currently that's somewhat painful, involving setting up a player and handlers
void readHandler (const string name, RoadmapRos* msg, ros::Time t, ros::Time shift, void* p)
{
  Roadmap* g = static_cast<Roadmap*> (p);
  *g = message::roadmapFromRos(*msg);
}



namespace message
{


/************************************************************
 * Conversion to and from ros messages
 ************************************************************/

Roadmap roadmapFromRos (const RoadmapRos& roadmap_ros)
{
  Roadmap r;

  // Nodes
  foreach (const Node& n, roadmap_ros.nodes) 
    r.addNode(n.position, n.id);

  // Edges
  foreach (const Edge& e, roadmap_ros.edges) 
    r.addEdge(e.node1, e.node2, e.cost);
  
  return r;
}


RoadmapRos roadmapToRos (const Roadmap& roadmap)
{
  RoadmapRos m;

  foreach (const NodeId id, roadmap.allNodes()) {
    Node n;
    const Point& p = roadmap.nodePoint(id);
    n.id = id;
    n.position.x = p.x;
    n.position.y = p.y;
    n.position.z = 0.0;
    m.nodes.push_back(n);

    foreach (const NodeId neighbor, roadmap.neighbors(id)) {

      // Avoid adding twice
      if (id<neighbor) {

        Edge e;
        e.node1 = id;
        e.node2 = neighbor;
        e.cost = roadmap.getCost(id, neighbor);
        
        m.edges.push_back(e);
      }
    }
  }
  return m;
}



/************************************************************
 * File io using bags
 ************************************************************/

// We write to file by converting to a ros message, then serializing using a bag
void writeRoadmapToFile (const Roadmap& g, const string& filename)
{
  ROS_DEBUG_STREAM_NAMED("file", "Trying to write roadmap to " << filename);
  ros::record::Recorder recorder;
  if (!recorder.open(filename))
    throw FileOpenException (filename);

  const RoadmapRos m = roadmapToRos(g);
  recorder.record(DUMMY_TOPIC, m, ros::Time(0.0));

  recorder.close();
  ROS_DEBUG_STREAM_NAMED("file", "Successfully wrote roadmap to " << filename);
}


Roadmap readRoadmapFromFile (const string& filename)
{
  ROS_DEBUG_STREAM_NAMED("file", "Trying to read roadmap from " << filename);
  ros::record::Player player;
  if (!player.open(filename, ros::Time(0.0)))
    throw FileOpenException(filename);


  Roadmap g;
  player.addHandler<RoadmapRos>(DUMMY_TOPIC, &readHandler, &g);
  if (!player.nextMsg())
    throw FileFormatException(filename);

  ROS_DEBUG_STREAM_NAMED("file", "Successfully read roadmap from " << filename);
  return g;
}














} // namespace msg

} // namespace topological_map
