/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


/**
 * \file 
 * 
 * Ros node that updates and publishes roadmap
 */


#include <topological_map/incremental_roadmap.h>
#include <topological_map/roadmap_msg.h>
#include <topological_map/exception.h>
#include <costmap_2d/costmap_2d_ros.h>
#include <ros/ros.h>
#include <boost/bind.hpp>
#include <boost/ref.hpp>
#include <boost/foreach.hpp>
#include <boost/thread.hpp>
#include <iostream>
#include <fstream>

#define foreach BOOST_FOREACH

namespace topological_map
{

using boost::bind;
using boost::ref;
using ros::Duration;
using ros::WallDuration;
using ros::NodeHandle;
using ros::WallRate;
using costmap_2d::Costmap2DROS;
using costmap_2d::Costmap2D;

/************************************************************
 * Constants
 ************************************************************/

const double DEFAULT_PUBLICATION_RATE=.3;
const double DEFAULT_UPDATE_RATE=.2;
const double DEFAULT_EDGE_DISTANCE_THRESHOLD=2.0;
const double DEFAULT_ROADMAP_SAVE_RATE=.1;
const double EDGE_REMOVAL_MULTIPLIER=1.0;


/************************************************************
 * Helper types
 ************************************************************/

typedef boost::shared_mutex Mutex;

struct SharedLock
{
  SharedLock (const string& s, Mutex& m) : lock(m), s(s)
  { 
    ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Obtaining shared lock with " << s); 
  }
  ~SharedLock () { ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Releasing shared lock with " << s); }
  boost::shared_lock<Mutex> lock;
  string s;
};

struct UpgradableLock
{
  UpgradableLock (const string& s, Mutex& m) : lock(m), s(s)
  { 
    ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Obtaining upgradable lock with " << s); 
  }
  ~UpgradableLock () { ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Releasing upgradable lock with " << s); }
  boost::upgrade_lock<Mutex> lock;
  string s;
};

struct WriteLock
{
  WriteLock (const string& s, UpgradableLock& l) : lock(l.lock), s(s)
  { 
    ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Obtaining write lock with " << s); 
  }
  ~WriteLock () { ROS_DEBUG_STREAM_NAMED ("roadmap_server_lock", "Releasing write lock with " << s); }
  boost::upgrade_to_unique_lock<Mutex> lock;
  string s;
};



/************************************************************
 * Main class def
 ************************************************************/
                                                         
                                                           
/// \brief Node that generates and publishes roadmap
///
/// ROS API
/// Inputs
/// - costmap parameters in private namespace ~costmap (for static map)
/// - if the ~input_file parameter is provided, the initial roadmap is loaded from there (if it exists)
/// - if the ~output_file parameter is provided, roadmaps are written there
/// - ~update_rate governs how often the roadmap is updated
/// - ~publication_rate governs how often the current roadmap is published
/// Outputs
/// - publishes roadmap at rate ~publication_rate hz (default .3) on topic roadmap
/// - if the ~output_file parameter is provided, writes the roadmap there
class RoadmapServer
{
public:

  /// Initialize roadmap server based on ros params, then spawn
  /// update and publish loops and return
  RoadmapServer();

private:
  
  /****************************************
   * Ops
   ****************************************/

  /// The main update loop
  void updateMap ();

  /// Publish current roadmap 
  void publishRoadmap ();

  /// Search for uncovered cells and add waypoints to cover them
  void addWaypoints ();

  /// Search for redundant edges and remove them
  void pruneEdges ();
  
  /// Tell the incremental roadmap about new costmap
  void updateMetricMap ();

  /// Save the map to file if necessary
  void saveMap ();

  /****************************************
   * Fields
   ****************************************/
  
  // Node handle
  NodeHandle nh_;

  // Transform listener
  tf::TransformListener tf_;

  // Costmap (used to keep track of metric map)
  Costmap2DROS cmap_ros_;

  // Number governing when nodes are connected (roughly, typical edge length).  
  // Higher means graph has more edges.
  const double edge_distance_threshold_;

  // Mutex for access to roadmap 
  Mutex mutex_;

  // Underlying roadmap computation state
  IncrementalRoadmap roadmap_;

  // Publisher for maps
  const ros::Publisher roadmap_pub_;

  // Rate of roadmap publication
  const double publication_rate_;

  // Rate of roadmap update
  const double update_rate_;

  // File to save map to 
  const string output_file_;

  // Whether the input file has been read
  bool input_file_read_if_necessary_;

  // Thread for publication loop (pending fix to timer bug)
  const boost::thread publication_thread_;

  // Thread for update loop (pending fix to timer bug)
  const boost::thread update_thread_;
  
};


/************************************************************
 * Utilities
 ************************************************************/

Costmap2D getCostmap (Costmap2DROS* cmap_ros)
{
  Costmap2D cmap;
  cmap_ros->getCostmapCopy(cmap);
  return cmap;
}

double getPrivateParam(const string& name, const double default_val)
{
  double t;
  NodeHandle nh("~");
  nh.param(name, t, default_val);
  return t;
}

string getPrivateParam(const string& name, const string& default_val)
{
  string s;
  NodeHandle nh("~");
  nh.param(name, s, default_val);
  return s;
}


/************************************************************
 * Initialization
 ************************************************************/

RoadmapServer::RoadmapServer () : 
  tf_(Duration(10)), cmap_ros_("costmap", tf_), 
  edge_distance_threshold_(getPrivateParam("edge_distance_threshold", DEFAULT_EDGE_DISTANCE_THRESHOLD)),
  roadmap_(getCostmap(&cmap_ros_), edge_distance_threshold_),
  roadmap_pub_(nh_.advertise<RoadmapRos>("roadmap", 10)),
  publication_rate_(getPrivateParam("publication_rate", DEFAULT_PUBLICATION_RATE)),
  update_rate_(getPrivateParam("update_rate", DEFAULT_UPDATE_RATE)),
  output_file_(getPrivateParam("output_file", "")),
  input_file_read_if_necessary_(false),
  publication_thread_(&RoadmapServer::publishRoadmap, this),
  update_thread_(&RoadmapServer::updateMap, this)
{
  string input_file = getPrivateParam("input_file", "");
  if (!input_file.empty()) {
    UpgradableLock lock("read roadmap", mutex_);
    WriteLock upgraded_lock("read roadmap", lock);
    ROS_DEBUG_STREAM_NAMED ("roadmap_io", "Reading roadmap from " << input_file);
    try {
      Roadmap r = message::readRoadmapFromFile(input_file);
      roadmap_.setRoadmap(r); // Note that roadmap_ is an IncrementalRoadmap that contains a Roadmap object
    }
    catch (const FileOpenException& e) {
      ROS_DEBUG_STREAM_NAMED ("roadmap_io", "Couldn't open " << input_file << "; falling back to empty roadmap");
    }
  }
  input_file_read_if_necessary_=true;
}




/************************************************************
 * Ros interface
 ************************************************************/

// Publishes roadmap.
void RoadmapServer::publishRoadmap ()
{
  WallRate r(publication_rate_);
  while (ros::ok()) {
    {
      SharedLock lock("publisher", mutex_);
      roadmap_pub_.publish(message::roadmapToRos(roadmap_.roadmapSnapshot()));
    }
    r.sleep();
  }
}


/************************************************************
 * Map updates
 ************************************************************/

// The main map update 
void RoadmapServer::updateMap ()
{
  WallRate r(update_rate_);

  while (!input_file_read_if_necessary_) {}
  while (ros::ok()) {
    updateMetricMap ();
    addWaypoints();
    pruneEdges();
    saveMap();
    r.sleep();
  }
}


void RoadmapServer::updateMetricMap ()
{
  UpgradableLock lock("update metric map", mutex_);
  {
    WriteLock upgraded_lock("update metric map", lock);
    roadmap_.updateMetricMap (getCostmap(&cmap_ros_));
  }
}

void RoadmapServer::addWaypoints ()
{
  ROS_DEBUG_NAMED ("roadmap_server", "Adding waypoints");
  
  // Get parameters
  const unsigned nx = roadmap_.getXSize();
  const unsigned ny = roadmap_.getYSize();

  // Loop over grid (in stages for more fine-grained locking)
  const unsigned skip=50;
  for (unsigned x0=0; x0<nx; x0+=skip) {
    UpgradableLock lock("add waypoints", mutex_);
    for (unsigned x=x0; (x<x0+skip) && (x<nx); x++) {
      if (!ros::ok()) break;

      for (unsigned y=0; y<ny; ++y) {
        const Cell2D c(x,y);

        // Check if need to add a waypoint
        if (roadmap_.knownFree(c) && !roadmap_.isCovered(c)) {
          
          Cell2D c2 = roadmap_.bestNearbyCell(c);

          // If yes, obtain a write lock and add it
          {
            WriteLock upgraded_lock("add waypoints", lock);
            roadmap_.addNode(c2);
          }
        }
      }
    }
  }
}


void RoadmapServer::pruneEdges () 
{
  ROS_DEBUG_NAMED ("roadmap_server", "Pruning edges");
  UpgradableLock lock("prune edges", mutex_);
  Roadmap copy = roadmap_.roadmapSnapshot();
  foreach (const NodeId id, copy.allNodes()) {
    bool removed_edge=true;
    while (removed_edge) {

      double best_edge_removal_cost=edge_distance_threshold_*EDGE_REMOVAL_MULTIPLIER;
      bool best_found=false;
      NodeId best=-1;

      foreach (const NodeId id2, copy.neighbors(id)) {
        double cost = roadmap_.edgePruneCost (id, id2, &copy);
        ROS_DEBUG_STREAM_NAMED ("roadmap_prune", "Cost of removing edge between " << id << " and " << id2 << " is " << cost);
        if (cost < best_edge_removal_cost) {
          best_edge_removal_cost = cost;
          best = id2;
          best_found = true;
        }
      }
    
      if (best_found) {
        ROS_DEBUG_STREAM_NAMED ("roadmap_prune", "removing edge between " << id << " and " << best);
        WriteLock upgraded_lock("prune edges", lock);
        roadmap_.removeEdge(id, best);
        copy.removeEdge(id, best);
      }
      else {
        removed_edge = false;
        ROS_DEBUG_NAMED ("roadmap_prune", "Not removing any edges");
      }
    }
  }
}


/************************************************************
 * Saving/loading map
 ************************************************************/

void RoadmapServer::saveMap ()
{
  ROS_DEBUG_NAMED ("roadmap_server", "Saving map");
  if (!output_file_.empty()) {
    ROS_DEBUG_STREAM_NAMED ("roadmap_io", "Saving roadmap to " << output_file_);
    SharedLock lock("save map", mutex_);
    message::writeRoadmapToFile(roadmap_.roadmapSnapshot(), output_file_);
    ROS_DEBUG_STREAM_NAMED ("roadmap_io", "Done saving roadmap");
  }
  else 
    ROS_DEBUG_STREAM_NAMED ("roadmap_io", "Not saving roadmap");
}




} // end namespace topological_map




namespace tmap=topological_map;

int main (int argc, char** argv)
{
  ros::init (argc, argv, "roadmap_server");
  
  tmap::RoadmapServer node;

  ROS_INFO ("Roadmap server initialization complete");

  ros::MultiThreadedSpinner spinner(4);
  spinner.spin();

  return 0;
}
