(defpackage :top-plan-stub
    (:use :cl :roslisp :roslisp-utils :topological_map-srv)
  (:import-from
   :topological_map-msg
   :id-val
   :position-val)
   
  (:export
   :stub-planner))

(in-package :top-plan-stub)

(defun normalize (s)
  (etypecase s
    (string (string-upcase s))
    (symbol (string-upcase (symbol-name s)))))

(defun equal-names (s1 s2)
  (equal (normalize s1) (normalize s2)))

(defun get-blocked-edges-from (pos blocked-edges)
  (declare (fixnum pos) (type vector blocked-edges))
  (loop 
    for e across blocked-edges
    for bl = (other-node e pos)
    when bl
      collect bl))

(defun other-node (e pos)
  (declare (fixnum pos) (type ros-message e))
  (with-fields (node1 node2) e
    (cond
      ((eql pos node1) node2)
      ((eql pos node2) node1)
      (t nil))))

(defun get-actions (pos blocked graph)
  (declare (fixnum pos) (list blocked) (type ros-message graph))
  (with-fields (edges) graph
    (loop 
      for e across edges
      for n = (other-node e pos)
      when (and n (not (member n blocked)))
	collect n)))


(def-service-callback (service-handler GetTopologicalPlan) (graph pos goal blocked_edges)
  (let ((a (get-action pos blocked_edges graph goal)))
    (make-response :plan_found t :wait nil :waypoints (coerce (cons pos a) 'vector))))


(defun marker-code (s)
  (symbol-code 'visualization_msgs-msg:<Marker> s))

(defun get-node (id nodes)
  (find id nodes :key #'id-val))

(defun get-position (id nodes)
  (position-val (get-node id nodes)))

(defvar *markers* nil)

(defun visualize (graph node action)
  (with-fields (nodes) graph
    (with-fields (x y) (get-position node nodes)
      (let ((m (make-msg "visualization_msgs/Marker"
			 (frame_id header) "/map"
			 id node
			 ns "planning_stub"
			 type (marker-code :cube)
			 action (marker-code action)
			 (x scale) .5
			 (y scale) .5
			 (z scale) .5
			 (x position pose) x
			 (y position pose) y
			 (z position pose) 0.0
			 (r color) .8
			 (g color) .7
			 (a color) 1.0)))
	(publish "visualization_marker" m)))))

  

(defun remove-old-markers (graph)
  (while *markers* (visualize graph (pop *markers*) :delete)))

(defun get-action (pos blocked-edges graph goal)
  (let ((blocked (get-blocked-edges-from pos blocked-edges))
	(path (list pos)))
    (format t "~&Service called with~&  Pos: ~a~&  Goal: ~a~&  Blocked edges: ~a~&  Actions: ~a"
	    pos goal blocked (get-actions pos blocked graph))
    (loop
      (format t "~&Action? ")
      (finish-output)
      (let ((tokens (mapcar #'read-from-string (tokens (read-line)))))
	(remove-old-markers graph)
	(if (every #'numberp tokens)
	    (return tokens)
	    (with-fields (nodes) graph
	      (dbind (cmd &rest args) tokens
		(cond
		  ((equal-names cmd "x")
		   (setq path (nreconc args path)))
		  ((equal-names cmd "d")
		   (repeat (or (first args) 1)
		     (pop path)))
		  ((equal-names cmd "r")
		   (return (reverse path)))
		  ((equal-names cmd "s")
		   (pop path)
		   (setq path (nreconc args path)))))
	      (format t "~&Current path is ~a" (reverse path))
	      (dolist (id (reverse path))
		(if (get-node id nodes)
		    (let ((bl (get-blocked-edges-from id blocked-edges)))
		      (visualize graph id :add)
		      (push id *markers*)
		      (format t "~&At position ~a, blocked edges are ~a and actions are ~a" 
			      id bl (get-actions id bl graph)))
		    (ros-warn planning-stub "Waypoint ~a not found in graph" id)))))))
    ))

(defun stub-planner ()
  (format t "~&Stub planner.   The way this works is that the service call happens at some node, say 2, and you type in a list of numbers separated by spaces like 3 6 9, and the service call will return list of waypoints 2, 3, 6, 9.  

Visualization commands (these don't return anything from the service, just visualize and print info):
- 'x 9 4' will add 9 and 4 to the end of the current path (extend)
- 'd 3' will remove the last three nodes from the current path; just 'd' will remove the last node (delete)
- 's 9 4' is like doing 'd' followed by 'x 9 4' (switch)

Finally, 'r' returns the currently visualized path.

")
  (with-ros-node ("stub_planner" :spin t)
    (advertise "visualization_marker" "visualization_msgs/Marker" :latch t)
    (register-service-fn "get_topological_plan" #'service-handler 'GetTopologicalPlan)))
  
	  