/*
 * Copyright (c) 2008, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <topological_map/exception.h>
#include <topological_map/roadmap.h>
#include <topological_map/roadmap_msg.h>
#include <topological_map/grid_graph.h>
#include <topological_map/point_ops.h>
#include <topological_map/incremental_roadmap.h>
#include <topological_map/utils.h>
#include <iostream>
#include <sstream>
#include <cmath>
#include <costmap_2d/costmap_2d.h>
#include <gtest/gtest.h>
#include <boost/foreach.hpp>


/****************************************
 * FILEWIDE DEFINE
 ****************************************/
#define foreach BOOST_FOREACH

using namespace topological_map;
using namespace std;

using std::cout;
using std::endl;
using boost::extents;
using boost::shared_ptr;
using boost::tie;
using std::stringstream;
using topological_map::OccupancyGrid;
using boost::multi_array;
using costmap_2d::Costmap2D;

// Helpers

bool isNeighbor (const Roadmap& roadmap, const NodeId n1, const NodeId n2)
{
  return contains(roadmap.neighbors(n1), n2);
}


void setV (OccupancyGrid* grid, unsigned r0, unsigned dr, unsigned rmax, unsigned c0, unsigned dc, unsigned cmax, bool val) 
{
  for (unsigned r=r0; r<rmax; r+=dr) {
    for (unsigned c=c0; c<cmax; c+=dc) {
      (*grid)[r][c] = val;
    }
  }
}




// Tests



TEST(TopologicalMap, IncrementalRoadmap)
{
  OccupancyGrid grid(extents[22][25]);
  setV(&grid, 0, 1, 22, 0, 1, 25, false);
  setV(&grid, 0, 7, 22, 0, 1, 25, true);
  setV(&grid, 0, 1, 22, 0, 6, 25, true);
  grid[7][3] = false;
  Costmap2D cmap = gridToCostmap(grid);
  IncrementalRoadmap r(cmap, 3.0);
  const Roadmap& roadmap = r.roadmapSnapshot();

  const Cell2D c1(1,1);
  const Cell2D c2(4,4);
  const Cell2D c3(7,3);
  const Cell2D c4(10,10);
  const Cell2D c5(5,3);
  

  EXPECT_TRUE(!r.isCovered(c1));
  EXPECT_TRUE(!r.isCovered(c2));
  EXPECT_TRUE(!r.isCovered(c3));
  EXPECT_TRUE(!r.isCovered(c4));

  EXPECT_EQ(Cell2D(2,2), r.bestNearbyCell(c1));
  
  const NodeId n1 = r.addNode(Cell2D(2,2));
  
  EXPECT_TRUE(r.isCovered(c1));
  EXPECT_TRUE(!r.isCovered(c2));
  EXPECT_TRUE(!r.isCovered(c3));
  EXPECT_TRUE(!r.isCovered(c4));
  
  EXPECT_EQ(Cell2D(3,3), r.bestNearbyCell(c2));
  const NodeId n2 = r.addNode(c2);

  EXPECT_TRUE (isNeighbor(roadmap, n1, n2));

  EXPECT_TRUE(r.isCovered(c1));
  EXPECT_TRUE(r.isCovered(c2));
  EXPECT_TRUE(!r.isCovered(c3));
  EXPECT_TRUE(!r.isCovered(c4));

  const NodeId n3 = r.addNode(c3);
  const NodeId n4 = r.addNode(c4);

  EXPECT_TRUE (isNeighbor(roadmap, n1, n2));
  EXPECT_TRUE (isNeighbor(roadmap, n2, n3));
  EXPECT_TRUE (isNeighbor(roadmap, n1, n3));  
  EXPECT_TRUE (!isNeighbor(roadmap, n1, n4));  
  EXPECT_TRUE (!isNeighbor(roadmap, n2, n4));  
  EXPECT_TRUE (!isNeighbor(roadmap, n3, n4));  
  
  {
    Roadmap copy = roadmap;
    EXPECT_EQ(6u, r.edgePruneCost(n3, n2, &copy));
  }
  
  
  const NodeId n5 = r.addNode(c5);
  EXPECT_TRUE (isNeighbor(roadmap, n2, n5));
  EXPECT_TRUE (isNeighbor(roadmap, n3, n5));

  {
    Roadmap copy = roadmap;
    EXPECT_EQ(0u, r.edgePruneCost(n3, n2, &copy));
    EXPECT_EQ(4u, r.edgePruneCost(n3, n5, &copy));
    EXPECT_EQ(4u, r.edgePruneCost(n5, n2, &copy));
  }

  r.removeEdge(n2,n3);
  EXPECT_TRUE (isNeighbor(roadmap, n1, n2));
  EXPECT_TRUE (!isNeighbor(roadmap, n2, n3));
  EXPECT_TRUE (isNeighbor(roadmap, n1, n3));  
  
}

  


TEST(TopologicalMap, GridGraph)
{
  OccupancyGrid grid(extents[10][20]);
  grid[1][1] = true;
  grid[2][0] = true;
  GridGraph graph(grid);
  multi_array<unsigned, 2> distances = graph.obstacleDistances();
  EXPECT_EQ(distances[0][0], 1u);
  EXPECT_EQ(distances[0][1], 1u);
  EXPECT_EQ(distances[0][2], 1u);
  EXPECT_EQ(distances[0][3], 1u);
  EXPECT_EQ(distances[1][0], 1u);
  EXPECT_EQ(distances[1][1], 0u);
  EXPECT_EQ(distances[2][0], 0u);
  EXPECT_EQ(distances[2][1], 1u);
  EXPECT_EQ(distances[3][3], 4u);
  EXPECT_EQ(distances[1][4], 2u);
  vector<Cell2D> neighbors = graph.neighbors(Cell2D(3,3));
  EXPECT_EQ(4u, neighbors.size());
}

template <class T>
bool equalSets (const set<T>& s1, const set<T>& s2)
{
  bool ret=true;
  if (s1.size() != s2.size())
    ret=false;
  else {
    foreach (const T& x, s1)
      if (find(s2.begin(), s2.end(), x) == s2.end())
        ret=false;
  }
  if (!ret) {
    ROS_INFO ("S1:");
    foreach (const T& x, s1)
      ROS_INFO_STREAM (" " << x);
    ROS_INFO ("S2:");
    foreach (const T& x, s2)
      ROS_INFO_STREAM (" " << x);
  }
  return ret;
}

typedef set<NodeId> NodeSet;

TEST(TopologicalMap, Roadmap)
{
  // Checking the roadmap-ros conversion, and roadmap copy constructor
  Roadmap r;
  const NodeId n1 = r.addNode(makePoint(2.4, 1.2));
  const NodeId n2 = r.addNode(makePoint(1.6, 3.3));
  const NodeId n3 = r.addNode(makePoint(5.0, 2.2));
  const NodeId n4 = r.addNode(makePoint(1.1, 5632));
  
  r.addEdge(n1, n2, 2.2);
  r.addEdge(n2, n3, 11);
  r.addEdge(n2, n4, 31);
  r.addEdge(n4, n3, 235);
  

  NodeSet s1, s2, s3;
  s1.insert(n2);
  s2.insert(n1);
  s2.insert(n2);
  s2.insert(n3);
  EXPECT_TRUE (equalSets(s3, r.nodesWithinRadius(makePoint(3, 3), .5)));
  EXPECT_TRUE (equalSets(s1, r.nodesWithinRadius(makePoint(3, 3), 1.5)));
  EXPECT_TRUE (equalSets(s2, r.nodesWithinRadius(makePoint(3, 3), 3)));
  
  


  const RoadmapRos m = message::roadmapToRos(r);
  const Roadmap r2 = message::roadmapFromRos(m);
  
  const set<NodeId> nodes = r2.allNodes();
  EXPECT_EQ (4u, nodes.size());
  EXPECT_EQ (31, r2.getCost(n4, n2));
  EXPECT_TRUE (closeTo(2.2, r2.getCost(n1, n2)));
  const set<NodeId> neighbors3 = r2.neighbors(n3);
  EXPECT_EQ (2u, neighbors3.size());
  
  EXPECT_TRUE (contains(neighbors3, n2));
  EXPECT_TRUE (contains(neighbors3, n4));
  EXPECT_TRUE (!contains(neighbors3, n1));

  // const string test_file("/tmp/topological_map_test_saved_roadmap.bag");
  // message::writeRoadmapToFile (r2, test_file);
  // const Roadmap r3 = message::readRoadmapFromFile(test_file);
  // EXPECT_EQ(4u, r3.allNodes().size());
  // EXPECT_EQ(31, r3.getCost(n4, n2));
  // EXPECT_EQ(2u, r3.neighbors(n3).size());
}




int main (int argc, char** argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
