/*
 * Copyright (C) 2008, Willow Garage, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the names of Stanford University or Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include <ros/ros.h>
#include <topological_map/GetTopologicalPath.h>
#include <topological_map/ObservePlanFailed.h>
#include <ros/console.h>
#include <gtest/gtest.h>


using topological_map::GetTopologicalPath;
using topological_map::ObservePlanFailed;
using ros::Duration;

bool close_to(const double x, const double y)
{
  const double EPS = 1e-6;
  return abs(x-y)<EPS;
}


void check_reasonable_path(const GetTopologicalPath::Response& r, const unsigned l)
{
  unsigned n=r.waypoints.size();
  EXPECT_TRUE(n==l);
  EXPECT_TRUE(close_to(r.waypoints[0].x, 16.0));
  EXPECT_TRUE(close_to(r.waypoints[0].y, 28.0));
  EXPECT_TRUE(close_to(r.waypoints[n-1].x, 12.0));
  EXPECT_TRUE(close_to(r.waypoints[n-1].y, 2.0));
}

TEST(WaypointGraph, GetTopologicalPath)
{
  unsigned num_waits = 0;
  ros::NodeHandle nh;
  ros::ServiceClient client = nh.serviceClient<GetTopologicalPath>("get_topological_path");
  ros::ServiceClient plan_failed_client = nh.serviceClient<ObservePlanFailed>("observe_last_plan_failed");

  Duration d(5.0);

  while (!client.waitForExistence(d) ||
         !plan_failed_client.waitForExistence(d))
    ASSERT_TRUE (num_waits++ < 30);

  // There seems to be a race condition where sometimes the costmap isn't getting enough time
  // at startup, so the initial pose is unknown
  ROS_INFO ("waiting");
  d.sleep();
  
  GetTopologicalPath srv;
  srv.request.start.x = 16.0;
  srv.request.start.y = 28;
  srv.request.goal.x = 12;
  srv.request.goal.y = 2;
  srv.request.radius = 4;

  ASSERT_TRUE(client.call(srv));

  check_reasonable_path(srv.response, 19);

  ASSERT_TRUE(client.call(srv));
  check_reasonable_path(srv.response, 19);
  
  ObservePlanFailed fail_srv;
  fail_srv.request.pos.x = 15.0;
  fail_srv.request.pos.y = 27.0;
  fail_srv.request.blocked_edges.push_back(6);
  ASSERT_TRUE(plan_failed_client.call(fail_srv));
  
  ASSERT_TRUE(client.call(srv));
  check_reasonable_path(srv.response, 21);

  Duration(35.0).sleep();
  ASSERT_TRUE(client.call(srv));
  check_reasonable_path(srv.response, 19);

  ASSERT_TRUE(plan_failed_client.call(fail_srv));
  ASSERT_TRUE(client.call(srv));
  check_reasonable_path(srv.response, 21);

  fail_srv.request.blocked_edges[0]=5;
  ASSERT_TRUE(plan_failed_client.call(fail_srv));
  ASSERT_TRUE(client.call(srv));
  check_reasonable_path(srv.response, 19);
  
}


  



int main(int argc, char** argv)
{
  testing::InitGoogleTest(&argc, argv);
  ros::init(argc, argv, "waypoint_graph_tester");
  return RUN_ALL_TESTS();
}
